1024 => int WINDOW_SIZE; // window size
3.5 => float WAVEFORM_Y; // y position of waveform
-1.6 => float SPECTRUM_Y; // y position of spectrum
10 => float DISPLAY_WIDTH; // width of waveform and spectrum display
17 => int WATERFALL_DEPTH; // waterfall depth; number of lines deep
Math.PI / 2 => float radians; //rotate 90 degrees
8 => int NUM_RAYS;

// GG.fullscreen(); // uncomment to fullscreen
GG.camera() --> GGen dolly --> GG.scene(); // put camera on a dolly
GG.camera().posZ( 10 ); // position
GG.camera().clip( .05, WATERFALL_DEPTH * 10 ); // set clipping plane
GG.scene().backgroundColor( @(.5,.5,.5) ); // set bg color
GG.scene().disableFog();
GG.scene().light().ambient(@(1,1,1));
GG.scene().light().diffuse(@(1,1,1));

GCircle sun --> GG.scene();
sun.position( @(4.8, 3.5, -0.001) );
// sun.position( @(0, 0, -0.001) );
sun.scale( @(1, 1, 1) );
// sun.mat().color( @(1, 0.812, 0) );
FlatMaterial matSun;
FileTexture bgSun;
bgSun.path("sun.png");
bgSun.wrap(1, 1);
matSun.diffuseMap(bgSun);
sun.mat(matSun);

// waveform renderer for time as rays
// GLines sunraysBig[NUM_RAYS / 2];
// [@(1.5, 0), @(-1.5, 0), @(0, 1.5), @(0, -1.5)] @=> vec3 posRayBig[];
// 0 => int rIndexBig;
// for (auto r: sunraysBig) {
//     r --> sun;
//     r.mat().color(@(0.812, 0.812, 0));
//     r.scale(@(.1, .1));
//     r.position(posRayBig[rIndexBig]);
//     rIndexBig++;
// }

SndBuf ocean => dac; //sound buffer ugen lets us playback from file system
0.5 => ocean.gain;
me.dir() + "ocean.wav" => string filename;
filename => ocean.read;

fun void audio() {
    ocean.samples() / 10 => ocean.pos;
    1::ms => now;
}

GLines waveform -->  sun;
waveform.mat().color( @(0.812, 0.812, 0) ); //time domain color
waveform.scale(@(.1, .1));
waveform.position( @(1.5, 0) );

GLines waveform2 -->  sun;
waveform2.mat().color( @(0.812, 0.812, 0) ); //time domain color
waveform2.scale(@(.1, .1));
waveform2.position( @(-1.5, 0) );

GLines waveform3 -->  sun;
waveform3.mat().color( @(0.812, 0.812, 0) ); //time domain color
waveform3.scale(@(.1, .1));
waveform3.rotZ(radians);
waveform3.position( @(0, 1.5) );

GLines waveform4 -->  sun;
waveform4.mat().color( @(0.812, 0.812, 0) ); //time domain color
waveform4.scale(@(.1, .1));
waveform4.rotZ(radians);
waveform4.position( @(0, -1.5) );

GLines waveformSm1 -->  sun;
waveformSm1.mat().color( @(0.812, 0.812, 0) ); //time domain color
waveformSm1.scale(@(.05, .05));
waveformSm1.rotZ(Math.PI / 4);
waveformSm1.position( @(-1, 1) );

GLines waveformSm2 -->  sun;
waveformSm2.mat().color( @(0.812, 0.812, 0) ); //time domain color
waveformSm2.scale(@(.05, .05));
waveformSm2.rotZ(-1 * Math.PI / 4);
waveformSm2.position( @(1, 1) );

GLines waveformSm3 -->  sun;
waveformSm3.mat().color( @(0.812, 0.812, 0) ); //time domain color
waveformSm3.scale(@(.05, .05));
waveformSm3.rotZ(-5 * Math.PI / 4);
waveformSm3.position( @(-1, -1) );

GLines waveformSm4 -->  sun;
waveformSm4.mat().color( @(0.812, 0.812, 0) ); //time domain color
waveformSm4.scale(@(.05, .05));
waveformSm4.rotZ(-7 * Math.PI / 4);
waveformSm4.position( @(1, -1) );

// <<< GG.scene().light() >>>;

// make a waterfall
Waterfall waterfall --> GG.scene();
waterfall.posY( SPECTRUM_Y ); // translate down
// waterfall.rotZ(radians);
waterfall.scale(@(1.09, 1.09));

FlatMaterial mat;
FileTexture bg;
bg.path("bg.png");
bg.wrap(1, 1);
mat.diffuseMap(bg);
GPlane sky --> GG.scene();
sky.mat(mat);
sky.posZ(-10);

@(35,20) => sky.scale;


//don't change
adc => Gain input; // which input?
// SinOsc sine => Gain input => dac; .15 => sine.gain;
input => Flip accum => blackhole; // accumulate samples from mic
input => PoleZero dcbloke => FFT fft => blackhole; // take the FFT
.95 => dcbloke.blockZero; // set DC blocker
// set size of flip
WINDOW_SIZE => accum.size;
// set window type and size
Windowing.hann(WINDOW_SIZE) => fft.window;
// set FFT size (will automatically zero pad)
WINDOW_SIZE*2 => fft.size;
// get a reference for our window for visual tapering of the waveform
Windowing.hann(WINDOW_SIZE) @=> float window[];

//don't change
float samples[0]; // sample array
complex response[0]; // FFT response
vec3 positions[WINDOW_SIZE]; // a vector to hold positions

// custom GGen to render waterfall
class Waterfall extends GGen {
    0 => int playhead; // waterfall playhead
    GLines wfl[WATERFALL_DEPTH]; // lines
    @(0, 1, 1) => vec3 color; // color for each wave

    for( GLines w : wfl ) { // iterate over line GGens
        w --> this; // aww yea, connect as a child of this GGen
        w.mat().color(color); // color
    }

    // update position of each waterfall frequency
    fun void latest( vec3 positions[] ) {
        positions => wfl[playhead].geo().positions; // set line mesh's position
        playhead++; // advance playhead
        WATERFALL_DEPTH %=> playhead; // wrap it. mod playhead by waterfall depth and assign it to playhead
    }

    // 0 => int i;
    // 5.0 => float radius;
    // 0 => int circleX;
    // 0 => int circleY;
    // // vec3 pos[WINDOW_SIZE]; //position of our circle

    // radius + wfl[i] => float diameter;
    // (2 * Math.PI / WINDOW_SIZE) * i => float theta;    
    // circleX + diameter * Math.cos(theta) => positions.x;
    // circleY + diameter * Math.cos(theta) => positions.y;


    // update
    fun void update( float dt ) {
        playhead => int pos; // position
        // WATERFALL_DEPTH/2.0 => float thresh; // for color
        Math.randomf() * 8.5 => float thresh;

        WATERFALL_DEPTH - thresh => float fadeChunk; // depth
        
        for( int i; i < wfl.size(); i++ ) {
            // start with playhead-1 and go backwards
            pos--; if( pos < 0 ) WATERFALL_DEPTH-1 => pos;
            // offset Z
            wfl[pos].posY( -i * 0.15 );
            if( i > thresh ) {
                // wfl[pos].mat().color( @((fadeChunk-(i-thresh) / fadeChunk ), 1, 1) );
                wfl[pos].mat().color( @((fadeChunk / i ), 1, 1) );
                // <<< thresh + " " + (fadeChunk / i) >>>;
            } else {
                wfl[pos].mat().color( color );
            }
        }
    }
}

// map audio buffer to 3D positions
fun void map2waveform( float in[], vec3 out[] ) {
    if( in.size() != out.size() )    {
        <<< "size mismatch in map2waveform()", "" >>>;
        return;
    }
    // mapping to xyz coordinate
    int i;
    DISPLAY_WIDTH => float width;
    for( auto s : in )    {
        // space evenly in X
        -width/2 + width/WINDOW_SIZE*i => out[i].x;
        // map y, using window function to taper the ends
        s*6 * window[i] => out[i].y;
        // a constant Z of 0
        0 => out[i].z;
        // increment
        i++;
    }
}

// map FFT output to 3D positions
fun void map2spectrum( complex in[], vec3 out[] ) {
    if( in.size() != out.size() ) {
        <<< "size mismatch in map2spectrum()", "" >>>;
        return;
    }
    // mapping to xyz coordinate
    int i;
    DISPLAY_WIDTH => float width;
    for( auto s : in ) {
        // space evenly in X
        -width/2 + width/WINDOW_SIZE*i => out[i].x;
        // map frequency bin magnitide in Y        
        5 * Math.sqrt( (s$polar).mag * 25 ) => out[i].y;
        // constant 0 for Z
        0 => out[i].z;
        // increment
        i++;
    }
    waterfall.latest( out );
}

// do audio stuff
fun void doAudio() {
    while( true ) {
        // upchuck to process accum
        accum.upchuck();
        // get the last window size samples (waveform)
        accum.output( samples );
        // upchuck to take FFT, get magnitude reposne
        fft.upchuck();
        // get spectrum (as complex values)
        fft.spectrum( response );
        // jump by samples
        WINDOW_SIZE::samp/2 => now;
    }
}
spork ~ doAudio();

// fun void controlSine( Osc s )
// {
//     while( true )
//     {
//         100 + (Math.sin(now/second*5)+1)/2*20000 => s.freq;
//         10::ms => now;
//     }
// }
// spork ~ controlSine( sine );

spork ~ audio();

0 => int skyMov;
0.00015 => float skyRot;


// graphics render loop
while( true )
{
    // map to interleaved format
    map2waveform( samples, positions );

    waveform.geo().positions( positions );
    waveform2.geo().positions( positions );
    waveform3.geo().positions( positions );
    waveform4.geo().positions( positions );
    waveformSm1.geo().positions( positions );
    waveformSm2.geo().positions( positions );
    waveformSm3.geo().positions( positions );
    waveformSm4.geo().positions( positions );

    // for (auto r: sunraysBig) {
    //     r.geo.positions( positions );
    // }
    // sunraysBig[1].geo.positions(positions);

    sun.rotZ(0.01);

    if (skyMov == 150) {
        -1* skyRot => skyRot;
        sky.rotZ(skyRot);
        0 => skyMov;
    } else {
        sky.rotZ(skyRot);
    }
    skyMov++;
    // map to spectrum display
    map2spectrum( response, positions );
    // next graphics frame
    GG.nextFrame() => now;
}
